@@ -1,10 +1,10 @@
 ;;; sql-indent.el --- Support for indenting code in SQL files. -*- lexical-binding: t -*-
 
-;; Copyright (C) 2015, 2017  Free Software Foundation, Inc
+;; Copyright (C) 2015, 2017-2018  Free Software Foundation, Inc
 
-;; Author: Alex Harsanyi (AlexHarsanyi@gmail.com)
+;; Author: Alex Harsanyi <AlexHarsanyi@gmail.com>
 ;; Created: 27 Sep 2006
-;; Version: 1.0
+;; Version: 1.1
 ;; Keywords: languages sql
 ;; Homepage: https://github.com/alex-hhh/emacs-sql-indent
 ;;
@@ -23,19 +23,15 @@
 
 ;;; Commentary:
 ;;
-;; Add syntax-based indentation when editing SQL code: TAB indents the current
-;; line based on the syntax of the SQL code on previous lines.
-;; `sqlind-minor-mode' is a minor mode that enables/disables this
-;; functionality. To setup syntax-based indentation for every SQL buffer, add
-;; `sqlind-minor-mode' to `sql-mode-hook'.
+;; `sqlind-minor-mode' is a minor mode that enables syntax-based indentation
+;; for `sql-mode' buffers: the TAB key indents the current line based on the
+;; SQL code on previous lines.  To setup syntax-based indentation for every
+;; SQL buffer, add `sqlind-minor-mode' to `sql-mode-hook'.  Indentation rules
+;; are flexible and can be customized to match your personal coding style.
+;; For more information, see the "sql-indent.org" file.
 ;;
 ;; The package also defines align rules so that the `align' function works for
 ;; SQL statements, see `sqlind-align-rules'.
-;;
-;; Indentation rules can be customized, for more information, see the
-;; "README.md" file or the doc strings for `sqlind-basic-offset',
-;; `sqlind-default-indentation-offsets-alist' and
-;; `sqlind-indentation-syntax-symbols'.
 
 ;;; Code:
 
@@ -185,10 +181,15 @@ statement."
     (goto-char pos)
     (catch 'found
       (while (re-search-backward "," limit 'noerror)
-        (when (sqlind-same-level-statement (point) limit)
-          (forward-char 1)
-          (sqlind-forward-syntactic-ws)
-          (throw 'found (point)))))))
+	(when (sqlind-same-level-statement (point) limit)
+	  (forward-char 1)
+	  (sqlind-forward-syntactic-ws)
+	  (throw 'found (point))))
+      ;; nothing was found in (while ...) so try to find the first column definition.
+      (goto-char limit)
+      (forward-sexp)
+      (sqlind-forward-syntactic-ws)
+      (point))))
 
 (defun sqlind-syntax (context)
   "Return the most specific syntax of CONTEXT.
@@ -736,11 +737,40 @@ See also `sqlind-beginning-of-block'"
                 (progn (forward-word -1) (looking-at "\\<_begin\\_>")))
         (throw 'finished (list 'in-block 'exception))))))
 
+(defun sqlind-maybe-$$-statement ()
+  "If (point) is on a PostgreSQL $$ quote, report its syntax.
+
+PostgreSQL uses $$ to delimit SQL blocks in create statements,
+this function tries to determine if the $$ block is a begin or an
+end block and creates the appropiate syntactic context.
+
+See also `sqlind-beginning-of-block'"
+  (when (looking-at "\\$\\$")
+    (prog1 t
+      (let* ((saved-pos (point))
+             (previous-block (save-excursion
+                               (ignore-errors (forward-char -1))
+                               (cons (sqlind-beginning-of-block) (point))))
+             (previous-block-kind (nth 0 previous-block)))
+        (goto-char saved-pos)
+        (if (and (listp previous-block-kind)
+                 (eq (nth 0 previous-block-kind) 'defun-start))
+            (progn
+              (when (null sqlind-end-stmt-stack)
+                (throw 'finished (list 'in-begin-block 'defun (nth 1 previous-block-kind))))
+              (cl-destructuring-bind (pos kind _label) (pop sqlind-end-stmt-stack)
+                (unless (eq kind '$$)
+                  (throw 'finished
+                    (list 'syntax-error "bad closing for $$ begin block" (point) pos)))
+                (goto-char (cdr previous-block))))
+          ;; Assume it is an "end" statement
+          (push (list (point) '$$ "") sqlind-end-stmt-stack))))))
+
 (defconst sqlind-start-block-regexp
   (concat "\\(\\_<"
 	  (regexp-opt '("if" "then" "else" "elsif" "loop"
 			"begin" "declare" "create" "alter" "exception"
-			"procedure" "function" "end" "case") t)
+			"procedure" "function" "end" "case" "$$") t)
 	  "\\_>\\)\\|)")
   "Regexp to match the start of a block.")
 
@@ -750,6 +780,25 @@ See also `sqlind-beginning-of-block'"
 END-STATEMENT-STACK contains a list of \"end\" syntaxes in
 reverse order (a stack) and is used to skip over nested blocks."
   (interactive)
+  ;; This function works as follows: `sqlind-start-block-regexp` defines the
+  ;; keywords where it stops to inspect the code.  Each time it stops at one
+  ;; of these keywords, it checks to see if the keyword is inside a comment or
+  ;; string.  If the keyworkd is not inside a comment or string, the
+  ;; `sqlind-maybe-*` functions are called to check if the keyword is valid.
+  ;; Each of these functions will do one of the following:
+  ;;
+  ;; * throw a syntax object with a 'finished tag, if they decide that the
+  ;;   keyword is valid
+  ;;
+  ;; * return t to indicate that they decided that the keyword is not valid
+  ;;   and `sqlind-beginning-of-block` should search for the next keyword
+  ;;
+  ;; * return nil to indicate that they don't recognize the keyword and
+  ;;   another `sqlind-maybe-*` function should be called
+  ;;
+  ;; Some of these `sqlind-maybe-*` functions are specific to the
+  ;; `sql-product` and are only invoked for the speficied SQL dialect.
+  
   (catch 'finished
     (let ((sqlind-end-stmt-stack end-statement-stack))
       (while (re-search-backward sqlind-start-block-regexp sqlind-search-limit 'noerror)
@@ -765,6 +814,8 @@ reverse order (a stack) and is used to skip over nested blocks."
             (sqlind-maybe-begin-statement)
             (when (eq sql-product 'oracle) ; declare statements only start blocks in PL/SQL
               (sqlind-maybe-declare-statement))
+            (when (eq sql-product 'postgres)
+              (sqlind-maybe-$$-statement))
             (sqlind-maybe-create-statement)
             (sqlind-maybe-defun-statement))))
     'toplevel))
@@ -1320,7 +1371,8 @@ not a statement-continuation POS is the same as the
     ;; context
     ((and (memq syntax-symbol '(defun-start declare-statement toplevel
                                 package package-body))
-          (looking-at "begin\\_>"))
+          (or (looking-at "begin\\_>")
+              (and (eq sql-product 'postgres) (looking-at "\\$\\$"))))
      (push (cons (list 'block-start 'begin) anchor) context))
 
     ((and (memq syntax-symbol '(defun-start package package-body))
@@ -1362,7 +1414,15 @@ not a statement-continuation POS is the same as the
        (push (sqlind-refine-end-syntax
               (if (equal kind "") nil (intern kind))
               label (point) context)
-             context))))
+             context)))
+
+    ((and (eq sql-product 'postgres)
+          (not (eq syntax-symbol 'comment-continuation))
+          (looking-at "\\$\\$"))
+     (push (sqlind-refine-end-syntax
+            nil "" (point) context)
+           context))
+    )
   context))
 
 (defun sqlind-syntax-of-line ()
@@ -1940,7 +2000,7 @@ with AND, OR or NOT to be aligned so they sit left under the WHERE clause."
           "\\s-*")
   "Regexp to match a SQL expression operator.")
 
-(defun sqlind-adjust-operator (_syntax base-indentation)
+(defun sqlind-adjust-operator (syntax base-indentation)
   "Adjust the indentation for operators in select clauses.
 
 If the line to be indented starts with an operator, the
@@ -1959,22 +2019,27 @@ example:
 This is an indentation adjuster and needs to be added to the
 `sqlind-indentation-offsets-alist`"
   (save-excursion
-    (back-to-indentation)
-    ;; If there are non-word constituents at the beginning if the line,
-    ;; consider them an operator and line up the line to the first word of the
-    ;; line, not the operator
-    (cond ((looking-at sqlind-operator-regexp)
-	   (let ((ofs (length (match-string 0))))
-	     (forward-line -1)
-	     (end-of-line)
-	     (sqlind-backward-syntactic-ws)
-             ;; Previous function leaves us on the first non-white-space
-             ;; character.  This might be a string terminator (') char, move
-             ;; the cursor one to the left, so 'forward-sexp' works correctly.
-             (ignore-errors (forward-char 1))
-	     (forward-sexp -1)
-	     (max 0 (- (current-column) ofs))))
-	  ('t base-indentation))))
+    (with-syntax-table sqlind-syntax-table
+      (back-to-indentation)
+      ;; If there are non-word constituents at the beginning if the line,
+      ;; consider them an operator and line up the line to the first word of the
+      ;; line, not the operator
+      (cond ((looking-at sqlind-operator-regexp)
+	     (let ((ofs (length (match-string 0))))
+	       (if (eq (sqlind-syntax-symbol syntax)
+		       'select-column-continuation)
+		   (goto-char (sqlind-column-definition-start
+			       (point) (sqlind-anchor-point syntax)))
+		 (forward-line -1)
+		 (end-of-line)
+		 (sqlind-backward-syntactic-ws)
+		 ;; Previous function leaves us on the first non-white-space
+		 ;; character.  This might be a string terminator (') char, move
+		 ;; the cursor one to the left, so 'forward-sexp' works correctly.
+		 (ignore-errors (forward-char 1))
+		 (forward-sexp -1))
+	       (max 0 (- (current-column) ofs))))
+	    ('t base-indentation)))))
 
 (defun sqlind-lone-semicolon (syntax base-indentation)
   "Indent a lone semicolon with the statement start.  For example:
